#!/usr/bin/env ruby
##    Copyright 2010 Ernest Turro
##
##    This program is free software; you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation; either version 2 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program; if not, write to the Free Software
##    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

require 'tempfile'

# beststratumfilter takes transcript IDs, insert sizes and number of mismatches,
# returns transcript IDs and insert sizes in lowest mismatch stratum, no. of mismatches in this stratum and number of elements filtered
def beststratumfilter transcripts, isizes, mismatches
  numfiltered=0
  minmm = mismatches.min
  transcripts2 = []
  isizes2 = []
  mismatches.each_index do |i| 
    if mismatches[i]==minmm then # keep only alignments in best mismatch stratum
      transcripts2 << transcripts[i]
      isizes2 << isizes[i]
    else
      numfiltered=numfiltered+1 
    end
  end
  return transcripts2, isizes2, minmm, numfiltered
end


# takes array of transcript IDs which are already within best mismatch stratum and removes those whose isize
# is more than deviation_threshold away from the expected isize, unless all are beyond the expected threshold,
# in which case all transcripts are kept
# duplicate transcripts are removed and, if $useisizefilter, isize closest to expected out of all dups used (otherwise random choice)

def isizedeviationfilter transcripts, isizes, expected, deviation_threshold
  unless $useisizefilter
    # remove dups and return with random choice of isize out of dups
    t2isize={}
    transcripts.each_index {|i|t2isize[transcripts[i]]=isizes[i]}
    return t2isize.keys, t2isize.values, 0
  end

  transcripts2=[]
  numberfiltered=0
  isizesfiltered=[]
  isizesnorm = isizes.map{|x| (x.abs - expected).abs}

  t2isizenorm={}
  transcripts.each_index do |i|
    if t2isizenorm[transcripts[i]].nil?
      t2isizenorm[transcripts[i]]=isizesnorm[i]
    else
      t2isizenorm[transcripts[i]]=[t2isizenorm[transcripts[i]] , isizesnorm[i] ].min
    end
  end
  transcripts=t2isizenorm.keys
  isizesnorm=t2isizenorm.values

  oneisizewithin=false
  isizesnorm.each{|x| if x <= deviation_threshold then oneisizewithin=true; break; end }
  if oneisizewithin
    isizesnorm.each_index do |i|
      if isizesnorm[i] <= deviation_threshold then
        transcripts2 << transcripts[i]
        isizesfiltered << isizes[i]
      else
        numberfiltered += 1
      end
    end
    return transcripts2, isizesfiltered, numberfiltered
  else
    return transcripts, isizes, 0
  end
end


####################
####################

tg_regexp = />(E\S+).*gene:(E\S+)/
t_rxp_ind = 1
g_rxp_ind = 2
custom_ref = false

if ARGV[0] == "-m"
  custom_ref = true
  ARGV.shift
  tg_regexp = Regexp.new ARGV.shift
  t_rxp_ind = ARGV.shift.to_i
  g_rxp_ind = ARGV.shift.to_i
end


# global variable
$useisizefilter=false
if ARGV[0] == "-i"
  $useisizefilter=true
  ARGV.shift
  expected_isize=ARGV.shift.to_i
  isize_deviation_thres=ARGV.shift.to_i
end

if ARGV.size != 2 then
  $stderr.puts "Usage: #{File.basename($0)} [-m tg_regexp t_ind g_ind] [-i expected_isize deviation_thres] sam_file cdna_file > hits_file"
  $stderr.puts "Generate hits file for use with `mmseq`."
  $stderr.puts ""
  $stderr.puts "Mandatory arguments:"
  $stderr.puts "  sam_file:  SAM file containing alignments to reference transcripts."
  $stderr.puts "  cdna_file: FASTA file used as reference to produce SAM file."
  $stderr.puts ""
  $stderr.puts "Arguments to optional flag -m for use with non-Ensembl cDNA FASTA files:"
  $stderr.puts "  tg_regexp: regular expression matching FASTA entries, where pairs of brackets"
  $stderr.puts "             are used to capture transcript and gene IDs. Default: \">(E\\S+).*gene:(E\\S+)\""
  $stderr.puts "  t_ind:     index of bracket pair that captures the transcript ID. Default: 1."
  $stderr.puts "  g_ind:     index of bracket pair that captures the gene ID. Default: 2."
  $stderr.puts ""
  $stderr.puts "Arguments to optional flag -i to filter out paired alignments with improbable insert sizes:"
  $stderr.puts "  expected_isize:  expected insert size (e.g. 200)"
  $stderr.puts "  deviation_thres: number of base pairs away (e.g. 50) from expected_isize beyond which an alignment is"
  $stderr.puts "                   removed if another alignment for the same read is within devation_thres"
  $stderr.puts "                   base pairs of expected_isize - applicable to reads within best mismatch stratum"
  $stderr.puts ""
  exit
end

if tg_regexp.source[0,1] != ">" then
  $stderr.puts "Error: tg_regexp must start with a \'>\' character."
  exit
end

if `which sort`.length==0 then
  $stderr.puts "Error: Unix `sort` required but not found."
  exit
end

SAM_FILENAME=ARGV[0]
CDNA_FILENAME=ARGV[1]

$stderr.puts "IMPORTANT NOTE: the aligner used to produce the SAM file must have been configured to output _all valid alignments_ (e.g. using the -a option in bowtie) for mmseq to work properly."

# read in fasta elements [tid,gid,seq]
$stderr.print "Parsing cDNA fasta file..."

fasta = []
File.open(CDNA_FILENAME) do |f|
  s = f.gets.chomp
  begin
    m = tg_regexp.match(s)
    tid = m[t_rxp_ind]
    gid = m[g_rxp_ind]
    seq = f.gets.chomp
    while(! f.eof? and (s = f.gets.chomp)[0,1] != '>') do seq = seq + s end
    fasta.push [tid, gid, seq]
  end while not f.eof? 
end

fasta.each {|t| puts "@TranscriptMetaData\t#{t[0]}\t#{t[2].length}\tfalse" }

# sort by gene ID 
fasta.sort!{|x,y| x[1] <=> y[1] }
curset=[fasta[0][0]]
curgene=fasta[0][1]
(0..fasta.length-2).each do |i|
  if(fasta[i+1][1] == curgene) then
    curset << fasta[i+1][0]
    next
  end
  puts "@GeneIsoforms\t" + curgene + "\t" + curset.join("\t")
  curset=[fasta[i+1][0]]
  curgene=fasta[i+1][1]
end
puts "@GeneIsoforms\t" + fasta[fasta.length-1][1] + "\t" + curset.join("\t")

# sort by sequence
fasta.sort!{|x,y| x[2] <=> y[2] }
curset=[fasta[0][0]]
curseq=fasta[0][2]
(0..fasta.length-2).each do |i|
  if(fasta[i+1][2] == curseq) then
    curset << fasta[i+1][0]
    next
  end
  puts "@IdenticalTranscripts\t" + curset.join("\t") if curset.length>1
  curset=[fasta[i+1][0]]
  curseq=fasta[i+1][2]
end
puts "@IdenticalTranscripts\t" + curset.join("\t") if curset.length>1

$stderr.print "done.\n"

$stderr.print "Parsing SAM file...\r"

tempout=Tempfile.new("mmseq")

outsidetopstratum=0
outsidedeviationthres=0

tg_regexp = Regexp.new(tg_regexp.source[1..-1]) # remove the initial '>'

File.open(SAM_FILENAME) do |fin|
  currentread=""
  currenttranscripts=[]
  currentmismatches=[]
  currentisizes=[]
  lineno=0
  while(!fin.eof?) do
    currentline = fin.gets
    if lineno % 1000000 == 0 and lineno > 1 then $stderr.print "Parsing SAM file...#{lineno/1000000} million alignments\r" end
    lineno += 1

    # skip comments
    if(currentline[0,1]=='@') then next end

    currentline = currentline.chomp.split("\t")

    # skip if the query sequence itself is unmapped or the mate is unmapped 
    if currentline[1].to_i & 0xC > 0 then next end

    # get the NM on current line
    currentline_nm=0
    currentline[11..-1].each {|x| unless x.match(/NM:i:/).nil? then currentline_nm = x[5,x.length-5].to_i; break; end }

    # get the isize on current line
    currentline_isize=currentline[8].to_i

    unless currentline[0] == currentread then
      unless currentread.empty?
        currenttranscripts, currentisizes, minmm, numstratfiltered = beststratumfilter currenttranscripts, currentisizes, currentmismatches
        currenttranscripts, currentisizes, numisizefiltered = isizedeviationfilter currenttranscripts, currentisizes, expected_isize, isize_deviation_thres
        tempout.print currenttranscripts.join(" ") + " " + currentisizes.join(" ") + " " + minmm.to_s + "\n"
        outsidetopstratum = outsidetopstratum+numstratfiltered
        outsidedeviationthres = outsidedeviationthres+numisizefiltered
      end
      currenttranscripts = [custom_ref ? tg_regexp.match(currentline[2])[t_rxp_ind] : currentline[2]]
      currentmismatches = [currentline_nm]
      currentisizes = [currentline_isize]
      currentread=currentline[0]
      tempout.print ">" + currentread + " "
    else
      ct = custom_ref ? tg_regexp.match(currentline[2])[t_rxp_ind] : currentline[2]
      ind = currenttranscripts.index ct
      if currentline[1].to_i & 0x2 > 0 && currenttranscripts[-1] == ct # if we found a mate in paired-end data
        # add mismatch amount to get total for both mates
        currentmismatches[-1] += currentline_nm
      else # otherwise add as new alignment
        currenttranscripts << ct
        currentmismatches << currentline_nm
        currentisizes << currentline_isize
      end
    end
  end
  # flush last alignments
  currenttranscripts, currentisizes, minmm, numstratfiltered = beststratumfilter currenttranscripts, currentisizes, currentmismatches
  currenttranscripts, currentisizes, numisizefiltered = isizedeviationfilter currenttranscripts, currentisizes, expected_isize, isize_deviation_thres
  tempout.print currenttranscripts.join(" ") + " " + currentisizes.join(" ") + " " + minmm.to_s + "\n"
  outsidetopstratum = outsidetopstratum+numstratfiltered
  outsidedeviationthres = outsidedeviationthres+numisizefiltered
end

tempout.close

$stderr.print "Parsing SAM file...done.                      \n"

$stderr.print "Sorting reads and merging alignments...\r" 

# so now we have a temporary file with lines that look like
# >IL6_3125:7:1:7:193 ENST00000395892 ENST00000375784 ENST00000375780 ENST00000375779 27 27 27 27 2
# >read_id            t1              t2              t3              t4              isize1...4 no.mismatches

`sort -o #{tempout.path} #{tempout.path}`

fot = File.open("newisizes", "w")

File.open(tempout.path) do |fin|
  currentline=""
  currentread=""
  currenttranscripts=[]
  currentmismatches=[]
  currentisizes=[]
  while(!fin.eof?)
    currentline=fin.gets.split(" ")
    nt = (currentline.length-2)/2 # number of current transcripts
    unless currentline[0] == currentread
      unless currentread.empty?
        currenttranscripts, currentisizes, minmm, numstratfiltered = beststratumfilter currenttranscripts, currentisizes, currentmismatches
        currenttranscripts, currentisizes, numisizefiltered = isizedeviationfilter currenttranscripts, currentisizes, expected_isize, isize_deviation_thres
        puts currentread, currenttranscripts
        fot.puts currentisizes
        outsidetopstratum = outsidetopstratum+numstratfiltered
        outsidedeviationthres = outsidedeviationthres+numisizefiltered
      end
      currentread, currenttranscripts = currentline[0], currentline[1,nt]
      currentmismatches = Array.new(nt, currentline[-1].to_i)
      currentisizes = currentline[1+nt,nt].map{|x| x.to_i}
    else
#      $stderr.puts currentread + " in multiple locations in SAM file"
      currenttranscripts += currentline[1,nt]
      currentmismatches += Array.new(nt, currentline[-1].to_i)
      currentisizes += currentline[1+nt,nt].map{|x| x.to_i}
    end
  end
  currenttranscripts, currentisizes, minmm, numstratfiltered = beststratumfilter currenttranscripts, currentisizes, currentmismatches
  currenttranscripts, currentisizes, numisizefiltered = isizedeviationfilter currenttranscripts, currentisizes, expected_isize, isize_deviation_thres
  puts currentread, currenttranscripts
  fot.puts currentisizes
  outsidetopstratum = outsidetopstratum+numstratfiltered
  outsidedeviationthres = outsidedeviationthres+numisizefiltered
end
fot.close
tempout.unlink

$stderr.print "Sorting reads and merging alignments...done.\n" 

$stderr.print "Removed #{outsidetopstratum} alignments that were outside lowest mismatch stratum.\n"

$stderr.print "Removed #{outsidedeviationthres} alignments that were beyond isize deviation threshold compared to other alignments of the same read.\n" if $useisizefilter

$stderr.puts "Now run `mmseq` on your output."

